/* ---------------------------------------------------------------------
Author:     H. Carter Edwards
		http://www.ticam.utexas.edu/~carter
		carter@ticam.utexas.edu

Copyright:  Copyright (C) 1996   H. Carter Edwards

Purpose:    Asynchronous message management for MPI

License:    See 'License.Gnu'

Mail:
	University of Texas at Austin
	TICAM, TAY 2.400
	2313 Speedway
	Austin, TX  78712

Disclaimer:

    MMPI comes with ABSOLUTELY NO WARRANTY; for details read
    the file License.Gnu that came with this library.
    This is free software, and you are welcome to redistribute it
    under certain conditions, i.e. read the file 'License.Gnu'.
--------------------------------------------------------------------- */
/* ---------------------------------------------------------------------
   Limitations:
     * These routines are NOT thread safe.
     * These routines assume one thread of execution per "processor".
--------------------------------------------------------------------- */

#ifndef __MANAGEMENT_FOR_NONTHREADED_MPI_APPLICATIONS__
#define __MANAGEMENT_FOR_NONTHREADED_MPI_APPLICATIONS__

#ifdef __cplusplus
extern "C" {
#endif

#include <stdio.h>
#include <mpi.h>

#ifdef __cplusplus
}
#include <iostream.h>
#endif

/* ------------------------------------------------------------------ */
/* ------------------------------------------------------------------
  A coordinated set of output files, one per processor.
*/

#ifdef __cplusplus
ostream & MMPI_Log_stream(void); // Log file in stream format
extern "C" {
#endif

/* Initialized the base name of the log file before the log file
   has been accessed, preferably immediately following the call
   to MPI_Init.
   This base name is suffixed with the rank of
   the processor within MPI_COMM_WORLD.
 */
   
int MMPI_Log_init( const char * const base_name );

int    MMPI_Log_file_d(void);
FILE * MMPI_Log_file(void);

void MMPI_Log_abort(
  const char * const who ,
  const char * const msg ,
  const int mpi_why );

void MMPI_Log_message(
  const char * const who ,
  const char * const msg );

/* ---------------------------------------------------------------------
  Synchronous operations:
    Initialize MMPI tag management and enable MMPI_Barrier for the
    given Communicator.
    If tag range is not specified default values will be used:
       TAG_MIN = 0x6000 = 24576
       TAG_MAX = 0x7fff = 32767  (8192 tags)
    Each MMPI tag is tracked with a single bit, thus the
    default ranges require a bit array spanning 1 Kbytes.
*/

int MMPI_Enable( MPI_Comm );

int MMPI_Enable_tag( MPI_Comm Comm , unsigned TAG_MIN ,  unsigned TAG_MAX );

void MMPI_Disable( MPI_Comm );

/* ---------------------------------------------------------------------
  Synchronous operation: verify that the communicator's global tags
  have been consistently specified and reserved.
*/

int MMPI_Tag_verify( MPI_Comm Comm );

/*
  1) Get and reserve a  'global' tag.
  2) Release a reserved 'global' tag.

    These are local operations; however, they should be
  called in a prescribed order to insure that the same
  tag is reserved on each processor.  Calling immediately
  following a 'barrier' operation is effective.
*/

int MMPI_Tag_get_global( MPI_Comm Comm , int * Tag );
int MMPI_Tag_rel_global( MPI_Comm Comm , int * Tag );

/*
  1) Get and reserve a  'local' tag.
  2) Release a reserved 'local' tag.

    These are also local operations.  This set of tags
  are assumed to be only locally significant.
*/

int MMPI_Tag_get_local( MPI_Comm Comm , int * Tag );
int MMPI_Tag_rel_local( MPI_Comm Comm , int * Tag );

/* ------------------------------------------------------------------ */
/* ------------------------------------------------------------------ * 
 * The receive handler routine specification.
 * ------------------------------------------------------------------ */

typedef int MMPI_Request_handler(
  void        * extra_state ,
  MPI_Request * request ,
  MPI_Status  * status );

#define MMPI_REQUEST_HANDLER_NULL  ((MMPI_Request_handler *) NULL)

/*
   MMPI_POST_HANDLER may be called for a persistent communication
   request which is in the inactive state.  MPI_POST_HANDLER may
   be repeatedly called for the same request, so long as the
   request is in the inactive state.
 */

int MMPI_Post_handler(
  MPI_Request          request ,     /* IN: */
  void               * extra_state , /* IN: */
  MMPI_Request_handler handler );    /* IN: */

/* Service handlers */

int MMPI_Serve_handler(void);

/*
  Wait for a request to complete.
  Handlers will be served while waiting.
  If the request is associated with a handler that handler
  will be called befor the call to MMPI_Wait returns.
 */

int MMPI_Wait( MPI_Request * , MPI_Status * );

/*
  Test if a request has completed.
  Pending handlers will be called.
 */

int MMPI_Test( MPI_Request * , int * , MPI_Status * );

/*
  Query how much wall-clock time (MPI_Wtime) has been consumed
  waiting for messages within MMPI_Wait.
 */

double MMPI_Wait_idle_time(void); /* Total wall-clock time spent waiting */

/* ------------------------------------------------------------------ */
/* ------------------------------------------------------------------ */
/* Continue to perform handled asynchronous operations
   while performing the synchronous barrier operation.
   The 'MMPI_Barrier' is enabled via 'MMPI_Init' or 'MMPI_Init_tag'.
 */

int MMPI_Barrier( MPI_Comm );

/* ------------------------------------------------------------------ */
/* ------------------------------------------------------------------ */
/*
   Packing and unpacking buffer support 
 */

typedef struct MMPI_Buf_type * MMPI_Buf ;

#define MMPI_BUF_NULL	((MMPI_Buf)NULL)

int MMPI_Buf_create( int , MPI_Comm , MMPI_Buf * );
int MMPI_Buf_copy( MMPI_Buf , MMPI_Buf * );
int MMPI_Buf_reset( int , MPI_Comm , MMPI_Buf * );
int MMPI_Buf_free( MMPI_Buf * );

int MMPI_Buf_capacity( MMPI_Buf , int * );
int MMPI_Buf_position( MMPI_Buf , int * );
int MMPI_Buf_pointer(  MMPI_Buf , void ** );
int MMPI_Buf_size( MMPI_Buf , int * );
int MMPI_Buf_comm( MMPI_Buf , MPI_Comm * );
int MMPI_Buf_remain( MMPI_Buf , int * );
int MMPI_Buf_verify( MMPI_Buf );

int MMPI_Buf_unpack(
  MMPI_Buf     inbuf ,
  void       * outbuf ,
  int          outcount ,
  MPI_Datatype datatype );

int MMPI_Buf_pack(
  void       * inbuf ,
  int          incount ,
  MPI_Datatype datatype ,
  MMPI_Buf   * outbuf );

/* After a MMPI_Buf receive has completed the resulting MPI_Status
   must be incorporated into the MMPI_Buf for correct unpacking.
 */

int MMPI_Buf_status( MMPI_Buf , MPI_Status * );

int MMPI_Buf_recv(  MMPI_Buf, int src , int tag , MPI_Status * );
int MMPI_Buf_send(  MMPI_Buf, int dest, int tag );
int MMPI_Buf_rsend( MMPI_Buf, int dest, int tag );
int MMPI_Buf_ssend( MMPI_Buf, int dest, int tag );
int MMPI_Buf_bsend( MMPI_Buf, int dest, int tag );

int MMPI_Buf_irecv(  MMPI_Buf, int src,  int tag, MPI_Request * );
int MMPI_Buf_isend(  MMPI_Buf, int dest, int tag, MPI_Request * );
int MMPI_Buf_irsend( MMPI_Buf, int dest, int tag, MPI_Request * );
int MMPI_Buf_issend( MMPI_Buf, int dest, int tag, MPI_Request * );
int MMPI_Buf_ibsend( MMPI_Buf, int dest, int tag, MPI_Request * );

int MMPI_Buf_recv_init(  MMPI_Buf, int src,  int tag, MPI_Request * );
int MMPI_Buf_send_init(  MMPI_Buf, int dest, int tag, MPI_Request * );
int MMPI_Buf_rsend_init( MMPI_Buf, int dest, int tag, MPI_Request * );
int MMPI_Buf_ssend_init( MMPI_Buf, int dest, int tag, MPI_Request * );
int MMPI_Buf_bsend_init( MMPI_Buf, int dest, int tag, MPI_Request * );

/* ------------------------------------------------------------------ */
/* ------------------------------------------------------------------ */
/*
  Dynamic producer-consumer messages are received on the destination
  processor and "consumed" by a consumer handler.  The producer-
  consumer message must be an MMPI packed buffer.

  An application must synchronously create its MMPI_Con objects.
  Once created messages may be arbitrarily and asynchronously
  sent between processors.
 */

typedef struct MMPI_Con_type * MMPI_Con ;

#define MMPI_CON_NULL	((MMPI_Con)NULL)
/*
  The consumer receives messages via the MPI_PACKED datatype,
  removes its own header information, and then passes
  the receive buffer along to the consumer handler.
  The buffer is unpacked as:
    MPI_Unpack( recv_buffer , recv_size , recv_position , ... , recv_comm );
 */

typedef int (*MMPI_Con_handler)(
  void   * extra_state , /* IN: Extra state of handler */
  int      source ,      /* IN: Message source */
  MMPI_Buf recv_buf );   /* IN: Received message */

/* Global synchronous operations:
    Create  'consumer' object
    Destroy 'consumer' object
 */

int MMPI_Con_create(
  MPI_Comm         comm ,         /* IN */
  void           * extra_state ,  /* IN */
  MMPI_Con_handler recv_handler , /* IN */
  MMPI_Con       * consumer );        /* OUT */

int MMPI_Con_free( MMPI_Con * ); /* IN OUT: Destroy consumer entity */

int MMPI_Con_comm( MMPI_Con , MPI_Comm * );
int MMPI_Con_func( MMPI_Con , MMPI_Con_handler * ); 
int MMPI_Con_data( MMPI_Con , void ** ); 
int MMPI_Con_verify( MMPI_Con );

/*
  Pack and send a producer-consumer message.
  + Pack message data via 'MPI_Con_pack', which calls 'MPI_Pack'.
  + Send message via 'MMPI_Con_send'.
  + Clear message via 'MMPI_Con_clear'.
    If the receiver is ready with buffer space an 'Rsend' is done,
    otherwise the send blocks until the receiver is ready.
 */

int MMPI_Con_init(
  MMPI_Con   consumer ,
  MMPI_Buf * buffer );

int MMPI_Con_send(
  MMPI_Buf buffer ,
  int      destination , /* IN: */
  MMPI_Con consumer );   /* IN OUT: */

/*
  Con entities allocate and maintain internal buffers.
  These buffers may be cleared and deleted,
  i.e. reset, to recover memory.
 */

int MMPI_Con_reset( MMPI_Con );

/* Does the consumer have any ongoing activity? */

int MMPI_Con_test( MMPI_Con , int destination , int * flag );
  
/* Wait until all ongoing activity has completed */

int MMPI_Con_wait( MMPI_Con , int destination );

/* ------------------------------------------------------------------ */

#ifdef __cplusplus
}
#endif

/* ------------------------------------------------------------------ */

#endif

