

/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  Ann Walther, State of Texas House Research Department
                                           ann.walther@house.state.tx.us

  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu
 
***************************************************************************/


#include "PLA.h"

static int pla_comm_create_1d_from_2d(MPI_Comm  comm2d,
                                      MPI_Comm  *comm1d);

static   MPI_Comm   PLA_GLOBAL_COMMUNICATOR=MPI_COMM_NULL;
static   MPI_Comm   PLA_GLOBAL_COMMUNICATOR_1D=MPI_COMM_NULL;
static   int        PLA_INITIALIZED=0;


int  PLA_Comm_1D_to_2D (MPI_Comm  comm1d,
                        int       rows,
                        int       cols,
                        MPI_Comm  *comm2d)
{
int
   dims[2],
   periods[2]={TRUE, TRUE},
   reorder=FALSE,
   error=0;

   dims[0] = rows; dims[1] = cols;          
   error+=MPI_Cart_create(comm1d, 2, dims, periods, 
                          reorder, comm2d);
   return (error);
}


int  PLA_Comm_1D_to_2D_ratio (MPI_Comm  comm1d,
                              double    ratio,
                              MPI_Comm  *comm2d)
/* 
   Want to find rows and colums such that 
              rows/columns = ratio
   or at least a close approxximation.  Noting that 
   columns=num_proc/rows and rearranging
              rows*rows  = ratio*num_proc
   So find rows such that 
         min |rows*rows - ratio*num_proc|
   and (num_proc%rows)=0
*/
{
int
   error = 0,
   num_proc,
   target_rows,
   minimum,
   temp,
   factor,
   rows;

   error+= MPI_Comm_size  (comm1d, &num_proc);

   rows    = num_proc;
   minimum = num_proc*num_proc;
   target_rows = ratio*(float)num_proc + .5; 
   for ( factor=1; factor<=num_proc ; factor++ ) {
      if ( num_proc%factor == 0) { /* this is an integer factor */
         temp = factor*factor - target_rows;
         if (temp<0) temp = -temp;
         if (temp<minimum) {
	     minimum = temp;
             rows    = factor;
         }
      }
   }
   error += PLA_Comm_1D_to_2D (comm1d,
                               rows,
                               num_proc/rows,
                               comm2d);

    return(error);
}

void PLA_Sum( PLA_DOUBLE_COMPLEX *in, PLA_DOUBLE_COMPLEX *inout,
             int *length, MPI_Datatype *datatype_ptr)
{
   int i;
   PLA_DOUBLE_COMPLEX c;

   for(i = 0; i < *length; ++i)
   {
      c.real = inout->real+in->real;
      c.imaginary = inout->imaginary+in->imaginary;
      *inout = c;
      in++;
      inout++;
   }
}

int  PLA_Init (MPI_Comm  comm)
{
int    
   size,
   dims[2]={0,0},
   periods[2]={TRUE, TRUE},
   reorder=FALSE,
   error=0;
   extern MPI_Op PLA_SUM;
   extern void PLA_Sum ();

if (PLA_INITIALIZED) {
   printf (" ERROR in PLA_Init\n");
   printf (" PLA_Init called more than one time. \n");
   return(1);
}
PLA_INITIALIZED=1;
if (MPI_COMM_NULL == comm) { 
   error+= MPI_Comm_size  (MPI_COMM_WORLD, &size);
   error+= MPI_Dims_create(size, 2, dims);
   error+= MPI_Cart_create(MPI_COMM_WORLD, 2, dims, 
                           periods, reorder, 
                          &PLA_GLOBAL_COMMUNICATOR);
}
else {
   MPI_Comm_dup (comm,&PLA_GLOBAL_COMMUNICATOR);
}

error+=pla_comm_create_1d_from_2d(PLA_GLOBAL_COMMUNICATOR,
                                 &PLA_GLOBAL_COMMUNICATOR_1D);
/* Next line replaces (temporarily) old PLA_API_begin() */
   MMPI_Enable(PLA_GLOBAL_COMMUNICATOR_1D);
/* Initial any MPI datatypes that we define for fortran 
   compatability but that MPI only defines for C. */
error+=PLA_MPI_datatypes ();
error+=MPI_Op_create( PLA_Sum, TRUE, &PLA_SUM);
return (error);
}

int PLA_Initialized (int *initialized) 
{
if (MPI_COMM_NULL == PLA_GLOBAL_COMMUNICATOR)
   *initialized = 0;
else
   *initialized = 1;
return(0);
}



int PLA_Base_comm (MPI_Comm *comm)
{
   *comm = PLA_GLOBAL_COMMUNICATOR;
   return (0);
}

int PLA_Base_comm_1d (MPI_Comm *comm)
{
   *comm = PLA_GLOBAL_COMMUNICATOR_1D;
   return (0);
}

int   PLA_Finalize ()
{
   int error=0, my_pe;
   extern int PLA_MEMORY_ALLOC_COUNT;
   char message [ 128 ];
   extern MPI_Op PLA_SUM;
   extern int PLA_MEMORY_SET_ASIDE_TYPE; /* What type is the scratch space */

   MPI_Comm_rank ( PLA_GLOBAL_COMMUNICATOR_1D, & my_pe );

/* Free the scratch space */
  if(PLA_HEAP == PLA_MEMORY_SET_ASIDE_TYPE)
  {
printf("Freeing the four pointers\n");
     pla_free(PLA_GLOBAL_OBJ_STRUCT);
     pla_free(PLA_GLOBAL_OBJ_INDEX);
     pla_free(PLA_GLOBAL_VIEWS_STRUCT);
     pla_free(PLA_GLOBAL_VIEWS_INDEX);
  }

   PLA_INITIALIZED=0;
   /* Old PLA_API_end() functionality begins */
   /* Commented out -- will not work on the Cray!!! */
   /* MMPI_Barrier(PLA_GLOBAL_COMMUNICATOR_1D);
      MMPI_Disable(PLA_GLOBAL_COMMUNICATOR_1D);  */
   /* Old PLA_API_end() functionality ends */
   if ( PLA_GLOBAL_COMMUNICATOR_1D==MPI_COMM_NULL )
      PLA_Log_abort ( "PLA_Finalize",
                      "Global communicator is equal to MPI_COMM_NULL prior to MPI_Comm_free", PLA_ERROR_CODE, 0 );

   error+=MPI_Comm_free (&PLA_GLOBAL_COMMUNICATOR_1D);
   error+=MPI_Comm_free (&PLA_GLOBAL_COMMUNICATOR);

   if ( error != 0 )
      PLA_Log_message ( "PLA_Finalize", "MPI_Comm_free returned error" );

   PLA_GLOBAL_COMMUNICATOR=MPI_COMM_NULL;
   PLA_GLOBAL_COMMUNICATOR_1D=MPI_COMM_NULL;
   MPI_Op_free(&PLA_SUM);
   if ( PLA_MEMORY_ALLOC_COUNT != 0 )
   {
      sprintf ( message,
                "Number of PLAPACK unfreed pointers = %d\n",
                PLA_MEMORY_ALLOC_COUNT );
      PLA_Log_message ( "PLA_Finalize", message );
   }

   return ( PLA_SUCCESS );
}


static int pla_comm_create_1d_from_2d(MPI_Comm  comm2d,
                               MPI_Comm  *comm1d)
{
int   
  error=0,
  my_cvect_index,
  dims[2], 
  periods[2], 
  coords[2];
  /*
     Create a 1D communicator.  This must be in column major order and
     created from the cartisian grid passed in the calling sequence.
     They will be ordered based on the number my_cvect_index.  So we
     just have to make sure it is increasing in column major order.
     */

  error = MPI_Cart_get(comm2d, 2, dims, periods, coords);
  if (error) {
    printf ("Error returned from MPI_Cart_get\n");
    return(1);
  }

  my_cvect_index = coords[1] * dims[0] + coords[0];
  error = MPI_Comm_split( comm2d, 0, my_cvect_index, comm1d);
  if (error) {
    printf ("Error returned from MPI_Comm_split\n");
    return(1);
  }
}

